#include "VPADeterminizer.h"

#include "automaton/common/State.h"
#include "alphabet/Symbol.h"
#include "alphabet/LabeledSymbol.h"
#include "label/Label.h"
#include "label/LabelSetLabel.h"
#include "label/LabelPairLabel.h"
#include "label/ObjectLabel.h"
#include "object/Object.h"
#include <std/set.hpp>
#include <iostream>

namespace determinize {

label::Label VPADeterminizer::packToDVPAStateLabel(std::set<std::pair<label::Label, label::Label>>&& data) {
	std::set<label::Label> res;
	for(auto&& subData : std::move(data)) {
		res.insert(label::Label(label::LabelPairLabel(std::move(subData))));
	}
	return label::Label(label::LabelSetLabel(std::move(res)));
}

std::set<std::pair<label::Label, label::Label>> VPADeterminizer::unpackFromDVPAStateLabel(const label::Label& data) {
	std::set<std::pair<label::Label, label::Label>> res;
	for (const auto& subData : static_cast<const label::LabelSetLabel&>(data.getData()).getData()) {
		res.insert(static_cast<const label::LabelPairLabel&>(subData.getData()).getData());
	}
	return res;
}

label::Label VPADeterminizer::packToDVPAStackSymbolLabel(std::pair<std::set<std::pair<label::Label, label::Label>>, alphabet::Symbol>&& data) {
	std::set<label::Label> res;
	for(auto&& subData : std::move(data.first)) {
		res.insert(label::Label(label::LabelPairLabel(std::move(subData))));
	}
	return label::Label(label::LabelPairLabel(std::make_pair(label::Label(label::LabelSetLabel(std::move(res))), label::Label(label::ObjectLabel(alib::Object(std::move(data.second.getData())))))));
}

std::pair<std::set<std::pair<label::Label, label::Label>>, alphabet::Symbol> VPADeterminizer::unpackFromDVPAStackSymbolLabel(const label::Label& data) {
	std::set<std::pair<label::Label, label::Label>> res;
	const label::Label& labelPairFirst = static_cast<const label::LabelPairLabel&>(data.getData()).getData().first;
	for (const auto& subData : static_cast<const label::LabelSetLabel&>(labelPairFirst.getData()).getData()) {
		res.insert(static_cast<const label::LabelPairLabel&>(subData.getData()).getData());
	}

	const label::Label& labelPairSecond = static_cast<const label::LabelPairLabel&>(data.getData()).getData().second;
	const alib::Object& object = static_cast<const label::ObjectLabel&>(labelPairSecond.getData()).getData();
	return std::make_pair(res, alphabet::Symbol(static_cast<const alphabet::SymbolBase&>(object.getData())));
}

void addRetTransition(const automaton::State& from, const alphabet::Symbol& input, const alphabet::Symbol& dvpdaSymbol, const automaton::State& to, automaton::VisiblyPushdownNPDA& deterministic) {
	deterministic.addState(from);
	deterministic.addState(to);
	deterministic.addStackSymbol(dvpdaSymbol);
	
	deterministic.addReturnTransition(from, input, dvpdaSymbol, to);
}

void retInitial(const automaton::State& state, const alphabet::Symbol& pdaSymbol, const alphabet::Symbol& input, const automaton::VisiblyPushdownNPDA& nondeterministic, automaton::VisiblyPushdownNPDA& deterministic) {
	std::set<std::pair<label::Label, label::Label>> S = VPADeterminizer::unpackFromDVPAStateLabel(state.getName());

	std::set<std::pair<label::Label, label::Label>> S1;
	for(const auto& entry : S) {
		const label::Label& q = entry.first;
		const label::Label& q2 = entry.second;

		for(const auto& transition : nondeterministic.getReturnTransitions()) {
			if(q2 != std::get<0>(transition.first).getName()) continue;
			if(input != std::get<1>(transition.first)) continue;
			if(nondeterministic.getBottomOfTheStackSymbol() != std::get<2>(transition.first)) continue;

			for(const auto& to : transition.second) {
				const label::Label& q1 = to.getName();
				S1.insert(std::make_pair(q, q1));
			}
		}
	}

	addRetTransition(state, input, pdaSymbol, automaton::State(VPADeterminizer::packToDVPAStateLabel(std::move(S1))), deterministic);
}

void ret(const automaton::State& state, const alphabet::Symbol& pdaSymbol, const alphabet::Symbol& input, const automaton::VisiblyPushdownNPDA& nondeterministic, automaton::VisiblyPushdownNPDA& deterministic) {
	std::set<std::pair<label::Label, label::Label>> S = VPADeterminizer::unpackFromDVPAStateLabel(state.getName());
	std::pair<std::set<std::pair<label::Label, label::Label>>, alphabet::Symbol> pdaSymbolUnpack = VPADeterminizer::unpackFromDVPAStackSymbolLabel(static_cast<const alphabet::LabeledSymbol&>(pdaSymbol.getData()).getLabel());
	const std::set<std::pair<label::Label, label::Label>>& S1 = pdaSymbolUnpack.first;

	std::set<std::pair<label::Label, label::Label>> update;

	for(const auto& transition : nondeterministic.getCallTransitions()) {
		if(pdaSymbolUnpack.second != std::get<1>(transition.first)) continue;

		const label::Label& q = std::get<0>(transition.first).getName();
		for(const auto& to : transition.second) {
			const label::Label q1 = to.first.getName();
			const alphabet::Symbol& gamma = to.second;

			for(const auto& entry : S) {
				if(q1 != entry.first) continue;
				const label::Label& q2 = entry.second;

				for(const auto& transition2 : nondeterministic.getReturnTransitions()) {
					if(q2 != std::get<0>(transition2.first).getName()) continue;
					if(input != std::get<1>(transition2.first)) continue;
					if(gamma != std::get<2>(transition2.first)) continue;

					for(const auto& to2 : transition2.second) {
						const label::Label& qI = to2.getName();

						update.insert(std::make_pair(q, qI));
					}
				}
			}
		}
	}

	std::set<std::pair<label::Label, label::Label>> S2;
	for(const auto& entry : S1) {
		const label::Label& q = entry.first;
		const label::Label& q3 = entry.second;
		for(const auto& entry2 : update) {
			if(q3 != entry2.first) continue;

			const label::Label& qI = entry2.second;

			S2.insert(std::make_pair(q, qI));
		}
	}

	addRetTransition(state, input, pdaSymbol, automaton::State(VPADeterminizer::packToDVPAStateLabel(std::move(S2))), deterministic);
}

void addCallTransition(const automaton::State& from, const alphabet::Symbol& input, const automaton::State& to, const alphabet::Symbol& dvpdaSymbol, automaton::VisiblyPushdownNPDA& deterministic) {
	deterministic.addState(from);
	deterministic.addState(to);
	deterministic.addStackSymbol(dvpdaSymbol);

	deterministic.addCallTransition(from, input, to, dvpdaSymbol);
}

std::set<label::Label> retrieveDSubSet(const std::set<std::pair<label::Label, label::Label>>& localOperation) {
	std::set<label::Label> id;
	for(const auto& entry : localOperation) {
		id.insert(entry.second);
	}
	return id;
}

std::set<std::pair<label::Label, label::Label>> createIdentity(std::set<label::Label>&& states) {
	std::set<std::pair<label::Label, label::Label>> id;
	for(const label::Label& state : states) {
		id.insert(std::make_pair(state, state));
	}
	return id;
}

void call(const automaton::State& state, const alphabet::Symbol& input, const automaton::VisiblyPushdownNPDA& nondeterministic, automaton::VisiblyPushdownNPDA& deterministic) {
	std::set<std::pair<label::Label, label::Label>> S = VPADeterminizer::unpackFromDVPAStateLabel(state.getName());

	std::set<label::Label> R = retrieveDSubSet(S);
	std::set<label::Label> R1;

	for(const label::Label& q : R) {
		for(const auto& transition : nondeterministic.getCallTransitions()) {
			if(q != transition.first.first.getName()) continue;
			if(input != transition.first.second) continue;

			for(const auto& to : transition.second) {
				const label::Label& q1 = to.first.getName();

				R1.insert(q1);
			}
		}
	}

	addCallTransition(state, input, automaton::State(VPADeterminizer::packToDVPAStateLabel(createIdentity(std::move(R1)))), alphabet::Symbol(alphabet::LabeledSymbol(VPADeterminizer::packToDVPAStackSymbolLabel(std::make_pair(std::move(S), input)))), deterministic);
}

void addLocalTransition(const automaton::State& from, const alphabet::Symbol& input, const automaton::State& to, automaton::VisiblyPushdownNPDA& deterministic) {
	deterministic.addState(from);
	deterministic.addState(to);

	deterministic.addLocalTransition(from, input, to);
}

void local(const automaton::State& state, const alphabet::Symbol& input, const automaton::VisiblyPushdownNPDA& nondeterministic, automaton::VisiblyPushdownNPDA& deterministic) {
	std::set<std::pair<label::Label, label::Label>> S = VPADeterminizer::unpackFromDVPAStateLabel(state.getName());
	std::set<std::pair<label::Label, label::Label>> S1;

	for(const auto& entry : S) {
		label::Label q = entry.first;
		label::Label q2 = entry.second;
		for(const auto& transition : nondeterministic.getLocalTransitions()) {
			if(q2 != transition.first.first.getName()) continue;
			if(input != transition.first.second) continue;

			for(const auto& to : transition.second) {
				label::Label q1 = to.getName();

				S1.insert(std::make_pair(q, q1));
			}
		}
	}
	
	addLocalTransition(state, input, automaton::State(VPADeterminizer::packToDVPAStateLabel(std::move(S1))), deterministic);
}

const automaton::State* existsDirtyState(const automaton::VisiblyPushdownNPDA& d) {
	for(const automaton::State& state : d.getStates()) {
		const label::Label& stateLabel = state.getName();
		
		for(const auto& transition : d.getCallTransitions()) {
			if(stateLabel != transition.first.first.getName()) continue;

			goto continue2;
		}
		return &state;

continue2: while(false); //TODO remove this
	}
	return NULL;
}

void localClosure(std::set<label::Label>& states, const std::set<label::Label>& oldStates, const automaton::VisiblyPushdownNPDA& d) {
	std::set<label::Label> newStates;

	for(const label::Label& state : oldStates) {
		for(const auto& transition : d.getLocalTransitions()) {
			if(!transition.second.count(automaton::State(state))) continue;

			if(!states.count(transition.first.first.getName())) {
				states.insert(transition.first.first.getName());
				newStates.insert(transition.first.first.getName());
			}
		}

		for(const auto& transition : d.getReturnTransitions()) {
			if(!transition.second.count(automaton::State(state))) continue;

			const alphabet::Symbol& popSymbol = std::get<2>(transition.first);
			if(popSymbol == d.getBottomOfTheStackSymbol()) continue;
			const label::Label& popSymbolLabel = static_cast<const alphabet::LabeledSymbol&>(popSymbol.getData()).getLabel();
			const label::Label& statePart = static_cast<const label::LabelPairLabel&>(popSymbolLabel.getData()).getData().first;

			if(!states.count(statePart)) {
				states.insert(statePart);
				newStates.insert(statePart);
			}
		}
	}

	if(newStates.empty()) return;
	localClosure(states, newStates, d);
}

std::pair<label::Label, alphabet::Symbol>* existsDirtyStateSymbol(const automaton::VisiblyPushdownNPDA& d) {
	for(const automaton::State& state : d.getStates()) {
		const label::Label& stateLabel = state.getName();

		std::set<label::Label> lc { stateLabel };
		localClosure(lc, std::set<label::Label>{stateLabel}, d); //intentional copy

		std::set<alphabet::Symbol> topSymbols;
		for(const label::Label& localState : lc) {
			for(const auto& transition : d.getCallTransitions()) {
				for(const auto& to : transition.second) {
					if(localState != to.first.getName()) continue;

					topSymbols.insert(to.second);
				}
			}

			if(d.getInitialStates().count(localState)) {
				topSymbols.insert(d.getBottomOfTheStackSymbol());
			}
		}

		for(const auto& transition : d.getReturnTransitions()) {
			if(stateLabel != std::get<0>(transition.first).getName()) continue;

			topSymbols.erase(std::get<2>(transition.first));
		}

		if(!topSymbols.empty()) return new std::pair<label::Label, alphabet::Symbol>(stateLabel, *topSymbols.begin());
	}
	return NULL;
}

std::set<label::Label> retrieveLabels(const std::set<automaton::State>& states) {
	std::set<label::Label> labels;
	for(const automaton::State& state : states) {
		labels.insert(state.getName());
	}
	return labels;
}

automaton::VisiblyPushdownNPDA VPADeterminizer::determinize(const automaton::VisiblyPushdownNPDA& n) {
	automaton::VisiblyPushdownNPDA d(n.getBottomOfTheStackSymbol());
	d.setCallInputSymbols(n.getCallInputAlphabet());
	d.setLocalInputSymbols(n.getLocalInputAlphabet());
	d.setReturnInputSymbols(n.getReturnInputAlphabet());
	
	label::Label initialLabel = VPADeterminizer::packToDVPAStateLabel(createIdentity(retrieveLabels(n.getInitialStates())));
	d.addState(automaton::State(initialLabel));
	d.addInitialState(automaton::State(initialLabel));
	
	for(;;) {
		std::pair<label::Label, alphabet::Symbol>* stateSymbol = existsDirtyStateSymbol(d);
		const automaton::State* state = existsDirtyState(d);
		if(stateSymbol != NULL) {
			for(const alphabet::Symbol& symbol : n.getReturnInputAlphabet()) {
				if(stateSymbol->second == d.getBottomOfTheStackSymbol()) {
					retInitial(stateSymbol->first, stateSymbol->second, symbol, n, d);
				} else {
					ret(stateSymbol->first, stateSymbol->second, symbol, n, d);
				}
			}
			delete stateSymbol; //TODO remove this...
		} else if(state != NULL) {
			for(const alphabet::Symbol& symbol : n.getLocalInputAlphabet()) {
				local(*state, symbol, n, d);
			}
			for(const alphabet::Symbol symbol : n.getCallInputAlphabet()) {
				call(*state, symbol, n, d);
			}
		} else {
			break;
		}
	}

	std::set<label::Label> finalLabels = retrieveLabels(n.getFinalStates());
	for(const automaton::State& state : d.getStates()) {
		const std::set<label::Label> labels = retrieveDSubSet(VPADeterminizer::unpackFromDVPAStateLabel(state.getName()));

		if(!std::empty_intersection(finalLabels, labels)) {
			d.addFinalState(state);
		}
	}
	
	return d;
}

}
